//! # Asynchronous Common Subset algorithm.
//!
//! The Asynchronous Common Subset protocol assumes a network of _N_ nodes that send signed
//! messages to each other, with at most _f_ of them malicious, where _3 f < N_. Handling the
//! networking and signing is the responsibility of the user: only when a message has been
//! verified to be "from node i" (e.g. using cryptographic signatures), it can be handed to the
//! `CommonSubset` instance.
//!
//! Each node proposes an element for inclusion. Under the above conditions, the protocol
//! guarantees that all correct nodes output the same set, consisting of at least _N - f_ of the
//! proposed elements.
//!
//! ## How it works
//!
//! * `CommonSubset` instantiates one `Broadcast` algorithm for each of the participating nodes.
//! At least _N - f_ of these - the ones whose proposer is not faulty - will eventually output
//! the element proposed by that node.
//! * It also instantiates Binary Agreement for each participating node, to decide whether
//! that node's proposed element should be included in the common set. Whenever an element is
//! received via broadcast, we input "yes" (`true`) into the corresponding `Agreement` instance.
//! * When _N - f_ `Agreement` instances have decided "yes", we input "no" (`false`) into the
//! remaining ones, where we haven't provided input yet.
//! * Once all `Agreement` instances have decided, `CommonSubset` returns the set of all proposed
//! values for which the decision was "yes".

use std::collections::{BTreeMap, BTreeSet, VecDeque};
use std::fmt::Debug;
use std::sync::Arc;

use agreement::{self, Agreement, AgreementMessage, AgreementStep};
use broadcast::{self, Broadcast, BroadcastMessage, BroadcastStep};
use fault_log::FaultLog;
use fmt::HexBytes;
use messaging::{DistAlgorithm, NetworkInfo, Step, TargetedMessage};
use rand::Rand;

error_chain!{
    types {
        Error, ErrorKind, ResultExt, CommonSubsetResult;
    }

    links {
        Agreement(agreement::Error, agreement::ErrorKind);
        Broadcast(broadcast::Error, broadcast::ErrorKind);
    }

    errors {
        MultipleAgreementResults
        NoSuchAgreementInstance
        NoSuchBroadcastInstance
    }
}

// TODO: Make this a generic argument of `CommonSubset`.
type ProposedValue = Vec<u8>;

/// Message from Common Subset to remote nodes.
#[derive(Serialize, Deserialize, Clone, Debug, Rand)]
pub enum Message<NodeUid: Rand> {
    /// A message for the broadcast algorithm concerning the set element proposed by the given node.
    Broadcast(NodeUid, BroadcastMessage),
    /// A message for the agreement algorithm concerning the set element proposed by the given
    /// node.
    Agreement(NodeUid, AgreementMessage),
}

/// The queue of outgoing messages in a `CommonSubset` instance.
#[derive(Deref, DerefMut)]
struct MessageQueue<NodeUid: Rand>(VecDeque<TargetedMessage<Message<NodeUid>, NodeUid>>);

impl<NodeUid: Clone + Debug + Ord + Rand> MessageQueue<NodeUid> {
    /// Appends to the queue the messages from `agr`, wrapped with `proposer_id`.
    fn extend_agreement(&mut self, proposer_id: &NodeUid, agr: &mut Agreement<NodeUid>) {
        let convert = |msg: TargetedMessage<AgreementMessage, NodeUid>| {
            msg.map(|a_msg| Message::Agreement(proposer_id.clone(), a_msg))
        };
        self.extend(agr.message_iter().map(convert));
    }

    /// Appends to the queue the messages from `bc`, wrapped with `proposer_id`.
    fn extend_broadcast(&mut self, proposer_id: &NodeUid, bc: &mut Broadcast<NodeUid>) {
        let convert = |msg: TargetedMessage<BroadcastMessage, NodeUid>| {
            msg.map(|b_msg| Message::Broadcast(proposer_id.clone(), b_msg))
        };
        self.extend(bc.message_iter().map(convert));
    }
}

/// Asynchronous Common Subset algorithm instance
pub struct CommonSubset<NodeUid: Rand> {
    /// Shared network information.
    netinfo: Arc<NetworkInfo<NodeUid>>,
    broadcast_instances: BTreeMap<NodeUid, Broadcast<NodeUid>>,
    agreement_instances: BTreeMap<NodeUid, Agreement<NodeUid>>,
    broadcast_results: BTreeMap<NodeUid, ProposedValue>,
    agreement_results: BTreeMap<NodeUid, bool>,
    /// Outgoing message queue.
    messages: MessageQueue<NodeUid>,
    /// The output value of the algorithm.
    output: Option<BTreeMap<NodeUid, ProposedValue>>,
    /// Whether the instance has decided on a value.
    decided: bool,
}

pub type CommonSubsetStep<NodeUid> = Step<NodeUid, BTreeMap<NodeUid, ProposedValue>>;

impl<NodeUid: Clone + Debug + Ord + Rand> DistAlgorithm for CommonSubset<NodeUid> {
    type NodeUid = NodeUid;
    type Input = ProposedValue;
    type Output = BTreeMap<NodeUid, ProposedValue>;
    type Message = Message<NodeUid>;
    type Error = Error;

    fn input(&mut self, input: Self::Input) -> CommonSubsetResult<CommonSubsetStep<NodeUid>> {
        debug!(
            "{:?} Proposing {:?}",
            self.netinfo.our_uid(),
            HexBytes(&input)
        );
        let fault_log = self.send_proposed_value(input)?;
        self.step(fault_log)
    }

    fn handle_message(
        &mut self,
        sender_id: &Self::NodeUid,
        message: Self::Message,
    ) -> CommonSubsetResult<CommonSubsetStep<NodeUid>> {
        let fault_log = match message {
            Message::Broadcast(p_id, b_msg) => self.handle_broadcast(sender_id, &p_id, b_msg)?,
            Message::Agreement(p_id, a_msg) => self.handle_agreement(sender_id, &p_id, a_msg)?,
        };
        self.step(fault_log)
    }

    fn next_message(&mut self) -> Option<TargetedMessage<Self::Message, Self::NodeUid>> {
        self.messages.pop_front()
    }

    fn terminated(&self) -> bool {
        self.messages.is_empty() && self.agreement_instances.values().all(Agreement::terminated)
    }

    fn our_id(&self) -> &Self::NodeUid {
        self.netinfo.our_uid()
    }
}

impl<NodeUid: Clone + Debug + Ord + Rand> CommonSubset<NodeUid> {
    pub fn new(netinfo: Arc<NetworkInfo<NodeUid>>, session_id: u64) -> CommonSubsetResult<Self> {
        // Create all broadcast instances.
        let mut broadcast_instances: BTreeMap<NodeUid, Broadcast<NodeUid>> = BTreeMap::new();
        for proposer_id in netinfo.all_uids() {
            broadcast_instances.insert(
                proposer_id.clone(),
                Broadcast::new(netinfo.clone(), proposer_id.clone())?,
            );
        }

        // Create all agreement instances.
        let mut agreement_instances: BTreeMap<NodeUid, Agreement<NodeUid>> = BTreeMap::new();
        for proposer_id in netinfo.all_uids() {
            agreement_instances.insert(
                proposer_id.clone(),
                Agreement::new(netinfo.clone(), session_id, proposer_id.clone())?,
            );
        }

        Ok(CommonSubset {
            netinfo,
            broadcast_instances,
            agreement_instances,
            broadcast_results: BTreeMap::new(),
            agreement_results: BTreeMap::new(),
            messages: MessageQueue(VecDeque::new()),
            output: None,
            decided: false,
        })
    }

    fn step(
        &mut self,
        fault_log: FaultLog<NodeUid>,
    ) -> CommonSubsetResult<CommonSubsetStep<NodeUid>> {
        Ok(Step::new(
            self.output.take().into_iter().collect(),
            fault_log,
        ))
    }

    /// Common Subset input message handler. It receives a value for broadcast
    /// and redirects it to the corresponding broadcast instance.
    pub fn send_proposed_value(
        &mut self,
        value: ProposedValue,
    ) -> CommonSubsetResult<FaultLog<NodeUid>> {
        if !self.netinfo.is_validator() {
            return Ok(FaultLog::new());
        }
        let uid = self.netinfo.our_uid().clone();
        // Upon receiving input v_i , input v_i to RBC_i. See Figure 2.
        self.process_broadcast(&uid, |bc| bc.input(value))
    }

    /// Receives a broadcast message from a remote node `sender_id` concerning a
    /// value proposed by the node `proposer_id`.
    fn handle_broadcast(
        &mut self,
        sender_id: &NodeUid,
        proposer_id: &NodeUid,
        bmessage: BroadcastMessage,
    ) -> CommonSubsetResult<FaultLog<NodeUid>> {
        self.process_broadcast(proposer_id, |bc| bc.handle_message(sender_id, bmessage))
    }

    /// Receives an agreement message from a remote node `sender_id` concerning
    /// a value proposed by the node `proposer_id`.
    fn handle_agreement(
        &mut self,
        sender_id: &NodeUid,
        proposer_id: &NodeUid,
        amessage: AgreementMessage,
    ) -> CommonSubsetResult<FaultLog<NodeUid>> {
        // Send the message to the local instance of Agreement
        self.process_agreement(proposer_id, |agreement| {
            agreement.handle_message(sender_id, amessage)
        })
    }

    /// Upon delivery of v_j from RBC_j, if input has not yet been provided to
    /// BA_j, then provide input 1 to BA_j. See Figure 11.
    fn process_broadcast<F>(
        &mut self,
        proposer_id: &NodeUid,
        f: F,
    ) -> CommonSubsetResult<FaultLog<NodeUid>>
    where
        F: FnOnce(&mut Broadcast<NodeUid>) -> Result<BroadcastStep<NodeUid>, broadcast::Error>,
    {
        let mut fault_log = FaultLog::new();
        let value = {
            let broadcast = self
                .broadcast_instances
                .get_mut(proposer_id)
                .ok_or(ErrorKind::NoSuchBroadcastInstance)?;
            let step = f(broadcast)?;
            fault_log.extend(step.fault_log);
            self.messages.extend_broadcast(&proposer_id, broadcast);
            if let Some(output) = step.output.into_iter().next() {
                output
            } else {
                return Ok(fault_log);
            }
        };
        self.broadcast_results.insert(proposer_id.clone(), value);
        let set_agreement_input = |agreement: &mut Agreement<NodeUid>| {
            if agreement.accepts_input() {
                agreement.input(true)
            } else {
                Ok(Step::default())
            }
        };
        self.process_agreement(proposer_id, set_agreement_input)?
            .merge_into(&mut fault_log);
        Ok(fault_log)
    }

    /// Callback to be invoked on receipt of the decision value of the Agreement
    /// instance `uid`.
    fn process_agreement<F>(
        &mut self,
        proposer_id: &NodeUid,
        f: F,
    ) -> CommonSubsetResult<FaultLog<NodeUid>>
    where
        F: FnOnce(&mut Agreement<NodeUid>) -> Result<AgreementStep<NodeUid>, agreement::Error>,
    {
        let mut fault_log = FaultLog::new();
        let value = {
            let agreement = self
                .agreement_instances
                .get_mut(proposer_id)
                .ok_or(ErrorKind::NoSuchAgreementInstance)?;
            if agreement.terminated() {
                return Ok(fault_log);
            }
            let step = f(agreement)?;
            fault_log.extend(step.fault_log);
            self.messages.extend_agreement(proposer_id, agreement);
            if let Some(output) = step.output.into_iter().next() {
                output
            } else {
                return Ok(fault_log);
            }
        };
        if self
            .agreement_results
            .insert(proposer_id.clone(), value)
            .is_some()
        {
            return Err(ErrorKind::MultipleAgreementResults.into());
        }
        debug!(
            "{:?} Updated Agreement results: {:?}",
            self.netinfo.our_uid(),
            self.agreement_results
        );

        if value && self.count_true() == self.netinfo.num_nodes() - self.netinfo.num_faulty() {
            // Upon delivery of value 1 from at least N − f instances of BA, provide
            // input 0 to each instance of BA that has not yet been provided input.
            for (uid, agreement) in &mut self.agreement_instances {
                if agreement.accepts_input() {
                    let step = agreement.input(false)?;
                    fault_log.extend(step.fault_log);
                    self.messages.extend_agreement(uid, agreement);
                    if let Some(output) = step.output.into_iter().next() {
                        if self.agreement_results.insert(uid.clone(), output).is_some() {
                            return Err(ErrorKind::MultipleAgreementResults.into());
                        }
                    }
                }
            }
        }
        self.try_agreement_completion();
        Ok(fault_log)
    }

    /// Returns the number of agreement instances that have decided "yes".
    fn count_true(&self) -> usize {
        self.agreement_results.values().filter(|v| **v).count()
    }

    fn try_agreement_completion(&mut self) {
        if self.decided || self.count_true() < self.netinfo.num_nodes() - self.netinfo.num_faulty()
        {
            return;
        }
        // Once all instances of BA have completed, let C ⊂ [1..N] be
        // the indexes of each BA that delivered 1. Wait for the output
        // v_j for each RBC_j such that j∈C. Finally output ∪ j∈C v_j.
        if self.agreement_results.len() < self.netinfo.num_nodes() {
            return;
        }
        debug!(
            "{:?} All Agreement instances have terminated",
            self.netinfo.our_uid()
        );
        // All instances of Agreement that delivered `true` (or "1" in the paper).
        let delivered_1: BTreeSet<&NodeUid> = self
            .agreement_results
            .iter()
            .filter(|(_, v)| **v)
            .map(|(k, _)| k)
            .collect();
        debug!("Agreement instances that delivered 1: {:?}", delivered_1);

        // Results of Broadcast instances in `delivered_1`
        let broadcast_results: BTreeMap<NodeUid, ProposedValue> = self
            .broadcast_results
            .iter()
            .filter(|(k, _)| delivered_1.contains(k))
            .map(|(k, v)| (k.clone(), v.clone()))
            .collect();

        if delivered_1.len() == broadcast_results.len() {
            debug!(
                "{:?} Agreement instances completed:",
                self.netinfo.our_uid()
            );
            for (uid, result) in &broadcast_results {
                debug!("    {:?} → {:?}", uid, HexBytes(&result));
            }
            self.decided = true;
            self.output = Some(broadcast_results)
        }
    }
}
